import argparse
import dataclasses

from . import Exporter, Metric

from prometheus_client import Gauge

from typing import Dict, Iterator


@dataclasses.dataclass
class MetricSandbox:
    labels: Dict[str, str]
    value: float


class ArbitraryExporter(Exporter):
    def create_args_parser(self, require_config: bool = True) -> argparse.ArgumentParser:
        parser = super(ArbitraryExporter, self).create_args_parser(require_config=False)

        parser.add_argument(
            '--gauge',
            required=False,
            metavar='NAME:LABEL1_NAME=LABEL1_VALUE,LABEL2_NAME=LABEL2_VALUE,...:VALUE',
            default=[],
            action='append',
            help='Add a gauge metric. Repeat LABEL_NAME[LABEL_VALUE]=VALUE to add metrics for each label and its values'
        )

        return parser

    def create_metric(self) -> Iterator[Metric]:
        for gauge_info in self.options.gauge:
            name, *values = gauge_info.split(':')

            sandboxes = []

            for value in values:
                sandbox = MetricSandbox(
                    labels={},
                    value=float('NaN')
                )

                sandboxes.append(sandbox)

                for label_info in value.split(','):
                    if '=' in label_info:
                        label_name, label_value = label_info.split('=', 1)

                        sandbox.labels[label_name.strip()] = label_value.strip()

                    else:
                        sandbox.value = float(label_info)

            label_names = list(set(sum(
                [list(sandbox.labels.keys()) for sandbox in sandboxes],
                []
            )))

            metric = Gauge(
                name.strip(), 'no help', label_names,
            )

            for sandbox in sandboxes:
                metric.labels(**sandbox.labels).set(sandbox.value)

            yield metric


def main() -> None:
    """
    Emit given arbitrary metrics.
    """

    ArbitraryExporter().run()


if __name__ == "__main__":
    main()



import argparse
import collections
import dataclasses
import re

import bs4
import requests
from prometheus_client import Gauge

from . import Exporter, Metric

from typing import cast, Any, DefaultDict, Dict, Iterator, List

JSONType = Dict[str, Any]


@dataclasses.dataclass
class JenkinsConnection:
    url: str
    username: str
    password: str


@dataclasses.dataclass
class NodeInfo:
    api: Dict[str, Any]
    config: bs4.Tag


@dataclasses.dataclass
class NodesInfo:
    api: List[Dict[str, Any]]
    nodes: Dict[str, NodeInfo]


class JenkinsExporter(Exporter):
    def parse_args(self) -> argparse.Namespace:
        options = super(JenkinsExporter, self).parse_args()

        self.jenkins = JenkinsConnection(
            url=self.config['jenkins']['master_url'],
            username=self.config['jenkins']['username'],
            password=self.config['jenkins']['password']
        )

        return options

    def _fetch_json(self, url: str) -> JSONType:
        return cast(
            JSONType,
            requests.get(
                f'{self.jenkins.url}/{url}',
                auth=(self.jenkins.username, self.jenkins.password)
            ).json()
        )

    def _fetch_xml(self, url: str) -> bs4.BeautifulSoup:
        return bs4.BeautifulSoup(
            requests.get(
                f'{self.jenkins.url}/{url}',
                auth=(self.jenkins.username, self.jenkins.password)
            ).content,
            'xml'
        )

    def fetch_data(self) -> None:
        self.data_nodes = NodesInfo(
            api=[],
            nodes={}
        )

        # Fetch summary data we can get for all nodes via JSON API.
        self.data_nodes.api = cast(
            List[Dict[str, Any]],
            self._fetch_json('/computer/api/json?depth=1')['computer']
        )

        # For each node, fetch its configuration, and extract its JSON API data from the global structure.
        for node_info in self.data_nodes.api:
            display_name = node_info['displayName']

            if display_name == 'master':
                continue

            self.data_nodes.nodes[display_name] = NodeInfo(
                api=node_info,
                config=self._fetch_xml(f'/computer/{display_name}/config.xml')
            )

        self.data_queue = self._fetch_json('/queue/api/json')

    def _create_label_count_metric(self) -> Metric:
        # We initialize values for `expected-labels` to zero, but we also want to add
        # entries for *unexpected* labels, which is easier with defaultdict.
        labels: DefaultDict[str, float] = collections.defaultdict(float)

        if self.data_fetch_failed:
            # When data fetch failed, set all metrics to NaN.
            for label in self.config['expected-labels']:
                labels[label] = float('nan')

        else:
            # Make sure to set default value for all labels.
            for label in self.config['expected-labels']:
                labels[label] = 0

            for display_name, node in self.data_nodes.nodes.items():
                node_labels_el = node.config.find('label')
                if node_labels_el is None:
                    self.logger.error(
                        f'Node {display_name} has no "label" configuration'
                    )
                    continue

                for label in node_labels_el.text.split(' '):
                    labels[label] += 1

        metric = Gauge(
            'node_labels',
            'Number of executors having a given label',
            ['label']
        )

        for node_label, count in labels.items():
            metric.labels(
                label=node_label
            ).set(count)

        return metric

    def _create_build_count_metric(self) -> Metric:
        # We initialize values for `expected-jobs` to zero, but we also want to add
        # entries for *unexpected* job, which is easier with defaultdict.
        running_jobs: DefaultDict[str, float] = collections.defaultdict(float)
        buildable_jobs: DefaultDict[str, float] = collections.defaultdict(float)
        blocked_jobs: DefaultDict[str, float] = collections.defaultdict(float)

        if self.data_fetch_failed:
            # When data fetch failed, set all metrics to NaN.
            for job in self.config['expected-jobs']:
                running_jobs[job] = float('nan')
                buildable_jobs[job] = float('nan')
                blocked_jobs[job] = float('nan')

        else:
            # Make sure to set default value for all jobs.
            for job in self.config['expected-jobs']:
                running_jobs[job] = 0
                buildable_jobs[job] = 0
                blocked_jobs[job] = 0

            # Running builds first
            for display_name, node in self.data_nodes.nodes.items():
                for executor in node.api['executors']:
                    if executor['currentExecutable'] is None:
                        continue

                    if 'url' not in executor['currentExecutable']:
                        self.logger.error(
                            f'Executor {display_name}:{executor["number"]} has no build URL'
                        )
                        continue

                    match = re.match(
                        fr'{self.jenkins.url}/job/(.+?)/\d+/',
                        executor['currentExecutable']['url']
                    )

                    if match:
                        running_jobs[match.group(1)] += 1

            # Queued builds next
            for queue_item in self.data_queue['items']:
                item_task = queue_item['task']
                item_class = item_task['_class']

                # These are known for not having project name
                if item_class == 'org.jenkinsci.plugins.workflow.support.steps.ExecutorStepExecution$PlaceholderTask':
                    self.logger.warn(
                        'task of type "ExecutorStepExecution$PlaceholderTask" does not specify project name'
                    )

                    continue

                # Let's make sure we get nice error when the project name's missing in queue items that should
                # specify it, at least as far as we can.
                if 'name' not in item_task:
                    raise Exception(f'task of type "{item_class}" does not specify project name')

                item_task_name = item_task['name']

                if item_class == 'hudson.model.Queue$BuildableItem':
                    buildable_jobs[item_task_name] += 1

                elif item_class == 'hudson.model.FreeStyleProject':
                    buildable_jobs[item_task_name] += 1

                elif item_class == 'hudson.model.Queue$BlockedItem':
                    blocked_jobs[item_task_name] += 1

                else:
                    self.logger.warn(
                        f'task of type "{item_class}" (task name {item_task_name}) was not recognized'
                    )

        metric = Gauge(
            'build_count',
            'Number of builds of a given job',
            # Simple "job" would be nicer, but that would collide with Prometheus' "job" label.
            ['jenkins_job', 'state']
        )

        for job_name, count in running_jobs.items():
            metric.labels(
                jenkins_job=job_name,
                state='running'
            ).set(count)

        for job_name, count in buildable_jobs.items():
            metric.labels(
                jenkins_job=job_name,
                state='buildable'
            ).set(count)

        for job_name, count in blocked_jobs.items():
            metric.labels(
                jenkins_job=job_name,
                state='blocked'
            ).set(count)

        for job_name in buildable_jobs.keys():
            metric.labels(
                jenkins_job=job_name,
                state='queued'
            ).set(buildable_jobs[job_name] + blocked_jobs[job_name])

        return metric

    def create_metric(self) -> Iterator[Metric]:
        yield self._create_label_count_metric()
        yield self._create_build_count_metric()


def main() -> None:
    """
    Fetches metrics from Jenkins and exports them to Prometheus format.
    """

    JenkinsExporter().run()


if __name__ == "__main__":
    main()



import argparse
import collections
import json
import re
import subprocess

from . import Exporter, Metric

from prometheus_client import Gauge

from typing import Any, Iterator, List


class OpenstackExporter(Exporter):
    def fetch_data(self) -> None:
        """
        Fetch a openstack data from server. ``self.data`` would become an object with following attributes:

        * ``limits: [{'Name': 'nameOfData', 'Value': 1024}, ...]``
        * ``instances: [{"Status": ..., "Name": ..., "Image": ..., "ID": ..., "Flavor": ..., "Networks": ...}, ...]``
        """

        self.data = argparse.Namespace()

        try:
            output = subprocess.Popen(
                self._get_command_openstack() + [
                    'limits',
                    'show',
                    '-f', 'json',
                    '--absolute',
                    '--reserved'
                ], stdout=subprocess.PIPE
            )

            self.data.limits = json.loads(output.communicate()[0])

            output = subprocess.Popen(
                self._get_command_openstack() + [
                    'server',
                    'list',
                    '-f', 'json'
                ], stdout=subprocess.PIPE
            )

            self.data.instances = json.loads(output.communicate()[0])

            output = subprocess.Popen(
                self._get_command_openstack() + [
                    'network',
                    'list',
                    '-f', 'json'
                ], stdout=subprocess.PIPE
            )

            networks = json.loads(output.communicate()[0])

            self.data.networks = {}

            for network in networks:
                if re.match(self.config['networks'], network['Name']) is None:
                    continue

                output = subprocess.Popen(
                    self._get_command_openstack() + [
                        'ip',
                        'availability',
                        'show',
                        '-f', 'json',
                        network['Name']
                    ], stdout=subprocess.PIPE
                )

                network_info = json.loads(output.communicate()[0])

                # Subnet information is presented as a plaintext, possibly with multiple lines,
                # one subnet per line. No JSON structure...
                for subnet_raw_info in network_info['subnet_ip_availability'].split('\n'):
                    subnet_raw_info = subnet_raw_info.strip()

                    subnet_info = dict(re.findall(r"(\S+)='(.*?|\S+)'", subnet_raw_info))

                    if subnet_info['ip_version'] != '4':
                        continue

                    self.data.networks[network['Name']] = {
                        'total': float(subnet_info['total_ips']),
                        'used': float(subnet_info['used_ips'])
                    }

        except Exception:
            # Do not log Exception details, becuase it contains secret informations like a Openstack password!
            self.logger.error(
                f"Phoebe Openstack: Can not fetch Openstack data."
            )
            raise

    def _get_command_openstack(self) -> List[str]:
        """
        It uses credentials from settings to create an Openstack client command-line.

        :returns: Shell command to login to Openstack
        """
        return [
            'openstack',
            f'--os-auth-url={self.config["client"]["os_auth_url"]}',
            f'--os-identity-api-version={self.config["client"]["os_identity_api_version"]}',
            f'--os-user-domain-name={self.config["client"]["os_user_domain_name"]}',
            f'--os-project-domain-name={self.config["client"]["os_project_domain_name"]}',
            f'--os-project-name={self.config["client"]["os_project_name"]}',
            f'--os-username={self.config["client"]["os_username"]}',
            f'--os-password={self.config["client"]["os_password"]}'
        ]

    def _extract_value(self, name: str) -> Any:
        """
        It extract value from a Openstack statistical data.

        :param name str: A name of specific value
        :returns: A value of a given name
        """
        return list(filter(lambda d: d['Name'] == name, self.data.limits))[0]['Value']

    def create_metric(self) -> Iterator[Metric]:
        """
        Generate Prometheus data for metrics. Definition of each metric is read from the configuration
        and yielded to the caller.
        """

        # Tenant usage metrics
        for metric_spes in self.config['metrics']:
            metric = Gauge(
                metric_spes['name'], metric_spes['help'], ['type'],
            )

            for metric_label in metric_spes['labels']:
                if self.data_fetch_failed:
                    value = float('NaN')

                else:
                    value = self._extract_value(metric_label['os-property'])
                    value *= float(metric_label.get('scale', 1))

                metric.labels(type=metric_label['type']).set(value)

            yield metric

        # Instance states metric
        state_counts = collections.Counter([
            instance['Status'].lower() for instance in self.data.instances
        ])

        metric = Gauge(
            'instance_states',
            'Number of instances in given state',
            ['state']
        )

        for state in self.config.get('expected-states', []):
            metric.labels(state=state.lower()).set(0.0)

        for state, count in state_counts.items():
            metric.labels(state=state).set(float(count))

        yield metric

        metric = Gauge(
            'network_ips',
            'Number of IP addresses in given network and state',
            ['network', 'type']
        )

        for network_name, network_info in self.data.networks.items():
            metric.labels(network=network_name.lower(), type='total').set(network_info['total'])
            metric.labels(network=network_name.lower(), type='used').set(network_info['used'])

        yield metric


def main() -> None:
    """
    Fetches metrics from Openstack and exports them to Prometheus format.
    """

    OpenstackExporter().run()


if __name__ == "__main__":
    main()



import argparse
import sys
import logging
import time

from prometheus_client import CollectorRegistry, generate_latest
from prometheus_client.metrics import MetricWrapperBase as Metric
from prometheus_client.metrics import Gauge
import ruamel.yaml
import stackprinter

from typing import Any, Dict, Iterator, List

SettingsType = Dict[str, Any]


class Exporter:
    def __init__(self) -> None:
        self.logger = logging.getLogger()
        self.data_fetch_failed = False

        now = time.time()

        self._start = now
        self._data_fetch_start = now
        self._data_fetch_end = now
        self._end = now

    def create_args_parser(self, require_config: bool = True) -> argparse.ArgumentParser:
        """
        Create and return a parser for command-line options. The parser is already set to accept
        common, shared options ``--config``, ``--export``, ``--log-file`` and ``--verbose``.
        """

        args_parser = argparse.ArgumentParser()

        args_parser.add_argument(
            "--config", "-c",
            required=require_config,
            metavar="FILE",
            help="Path to plugin's YAML config file."
        )

        args_parser.add_argument(
            "--export", "-e",
            required=True,
            metavar="FILE",
            help="Path to the output file. Use '-' for standard output instead of a file."
        )

        args_parser.add_argument(
            "--log-file", "-l",
            default="-",
            required=False,
            metavar="FILE",
            help="Path to the log file. Default value '-' is stderr."
        )

        args_parser.add_argument(
            "--verbose", "-v",
            required=False,
            action="store_true",
            help="Increase verbosity"
        )

        args_parser.add_argument(
            '--include-batch-metrics', '-B',
            action='store_true',
            required=False,
            help='If set, Phoebe will add metrics for timestamps marking exporter stages.'
        )

        args_parser.add_argument(
            '--batch-metric-started',
            required=False,
            default='phoebe_start_time_seconds',
            help='Name of the metric providing timestamp of exporter start.'
        )

        args_parser.add_argument(
            '--batch-metric-finished',
            required=False,
            default='phoebe_end_time_seconds',
            help='Name of the metric providing timestamp of exporter exit.'
        )

        args_parser.add_argument(
            '--batch-metric-fetch-started',
            required=False,
            default='phoebe_start_fetch_time_seconds',
            help='Name of the metric providing timestamp of data fetch start.'
        )

        args_parser.add_argument(
            '--batch-metric-fetch-finished',
            required=False,
            default='phoebe_end_fetch_time_seconds',
            help='Name of the metric providing timestamp of data fetch exit.'
        )

        return args_parser

    def parse_args(self) -> argparse.Namespace:
        """
        Parse command-line arguments.

        Common options ``--config``, ``--export``, ``--log-file`` and ``--verbose`` are consumed immediately.
        """

        parser = self.create_args_parser()

        self.options = options = parser.parse_args()

        self.config_filepath = options.config
        self.export_filepath = options.export

        # Setup up logging errors to file or on stderr. Default error output is printed on stderr.
        if options.log_file == '-':
            log_handler = logging.StreamHandler()
        else:
            log_handler = logging.FileHandler(options.log_file, encoding="utf-8")
        log_handler.setFormatter(
            logging.Formatter(
                '[%(asctime)s] [%(levelname)s] %(message)s', "%Y-%m-%d %H:%M:%S"
            )
        )
        self.logger.addHandler(log_handler)
        self.logger_verbose = options.verbose

        if self.options.config:
            try:
                with open(options.config) as f:
                    self.config = ruamel.yaml.YAML().load(f)
            except Exception as exc:
                self.logger.critical(f"Can not open phoebe config file '{options.config}' \n\t\t{exc}")

                sys.exit(1)

        else:
            self.config = {}

        return options

    def fetch_data(self) -> None:
        """
        This method is responsible for acquiring data the exporter needs to emit metrics.

        If this method raises an exception, ``data_fetch_failed`` property of the instance is set to ``True``.
        """

        pass

    def create_metric(self) -> Iterator[Metric]:
        """
        This method is expected to yield one metric every time is called, or return when exporter
        runs out of metrics to export.
        """

        # This method is supposed to be a generator, overriden by child classes. We want to provide a default
        # behavior, which would be an "empty" iterator - when called, this method should yield no entries,
        # iteration over it should end immediately. This is achieved by doing nothing and calling `return`
        # right away.
        #
        # But `return` alone is not enough - this function must be a generator, and generator must have `yield`.
        # So, adding `yield` - and the order is important, `return` must go first. Othwerwise, `yield` would
        # actually yielded an item, `None`, but as said above, this generator is supposed to be empty.
        #
        # So, `return` to quit the generator, and `yield` to "trick" Python to treating this method as
        # a generator. The `yield` is never used, all calls end with `return`, but it's present.
        return
        yield

    def _create_source_reachable_metric(self, value: int) -> Gauge:
        metric = Gauge(
            'phoebe_source_reachable',
            'Whether or not the data were fetched successfully'
        )

        metric.set(value)

        return metric

    def _create_batch_metrics(self) -> List[Gauge]:
        if not self.options.include_batch_metrics:
            return []

        metrics_info = [
            (self.options.batch_metric_started, 'Timestamp of the exporter start', self._start),
            (self.options.batch_metric_finished, 'Timestamp of the exporter exit', self._end),
            (self.options.batch_metric_fetch_started, 'Timestamp of the data fetch start', self._data_fetch_start),
            (self.options.batch_metric_fetch_finished, 'Timestamp of the exporter exit', self._data_fetch_end)
        ]

        metrics = [
            Gauge(name, help_string) for name, help_string, _ in metrics_info
        ]

        for metric, (_, _, value) in zip(metrics, metrics_info):
            metric.set(value)

        return metrics

    def collect_metrics(self) -> CollectorRegistry:
        """
        Main workhorse: calls ``create_metric`` as long as necessary, adds gained metrics to a registry,
        and returns the registry.
        """

        registry = CollectorRegistry()

        for metric in self.create_metric():
            registry.register(metric)

        registry.register(self._create_source_reachable_metric(0 if self.data_fetch_failed else 1))

        self._end = time.time()

        for metric in self._create_batch_metrics():
            registry.register(metric)

        return registry

    def export_metrics(self, registry: CollectorRegistry) -> None:
        exported = generate_latest(registry)

        if self.export_filepath == '-':
            sys.stdout.write(exported.decode('utf-8'))

        else:
            with open(self.export_filepath, 'w') as f:
                f.write(exported.decode('utf-8'))

    def run(self) -> None:
        self.parse_args()

        self._data_fetch_start = time.time()

        try:
            self.fetch_data()

        except Exception as exc:
            self.data_fetch_failed = True

            self.logger.error(f'Failed to fetch data: {exc}')
            if self.logger_verbose:
                self.logger.error(stackprinter.format_current_exception())

        finally:
            self._data_fetch_end = time.time()

        try:
            registry = self.collect_metrics()

        except Exception as exc:
            self.logger.critical(f'Failed to collect metrics: {exc}')
            if self.logger_verbose:
                self.logger.critical(stackprinter.format_current_exception())

            sys.exit(1)

        try:
            self.export_metrics(registry)

        except Exception as exc:
            self.logger.critical(f'Failed to export metrics: {exc}')
            if self.logger_verbose:
                self.logger.critical(stackprinter.format_current_exception())

            sys.exit(1)
